import numpy as np

from pastasoss.soss_traces import rotate
from pastasoss.soss_traces import get_reference_trace
from pastasoss.soss_traces import REFERENCE_TRACE_FILES

from pastasoss.wavecal import get_wavecal_meta_for_spectral_order


def test_load_wavecal_model_order1():
    """Test if wavecol model json files are loaded in correctly and properly
    stored in its datamodel object for order 1"""
    # polynomial expected number of coefficient and degree
    NCOEF = 20
    POLY_DEGREE = 5
    INTERCEPT = 2.761308747854264  # in units of microns

    # fitted x-pixel columns limits
    X_MIN_SCALER = 73.62893071613087
    X_MAX_SCALER = 2039.7420006352984

    # fitted PWCPOW offset (i.e., PWCPOS-245.76)
    PWCPOS_OFFSET_MIN_SCALER = -0.1035520000001781
    PWCPOS_OFFSET_MAX_SCALER = 0.1628820800781341

    # load order 1 wavecal model
    order1_metadata = get_wavecal_meta_for_spectral_order("order1")

    assert order1_metadata.order == "order1"
    assert len(order1_metadata.coefficients) == NCOEF
    assert order1_metadata.poly_degree == POLY_DEGREE
    assert order1_metadata.intercept == INTERCEPT
    assert order1_metadata.scaler_data_min_[0] == X_MIN_SCALER
    assert order1_metadata.scaler_data_max_[0] == X_MAX_SCALER
    assert order1_metadata.scaler_data_min_[1] == PWCPOS_OFFSET_MIN_SCALER
    assert order1_metadata.scaler_data_max_[1] == PWCPOS_OFFSET_MAX_SCALER


def test_load_wavecal_model_order2():
    """Test if wavecol model json files are loaded in correctly and properly
    stored in its datamodel object for order 2"""

    # polynomial expected number of coefficient and degree
    NCOEF = 9
    POLY_DEGREE = 3
    INTERCEPT = 1.418562550851081  # in units of microns

    # fitted x-pixel columns limits
    X_MIN_SCALER = 0.0
    X_MAX_SCALER = 2048.0

    # fitted PWCPOW offset (i.e., PWCPOS-245.76)
    # this has been changed for order 2 to be
    # within in the allowed tolerance range.
    # order 1 will be treated the same later
    PWCPOS_OFFSET_MIN_SCALER = 245.5929
    PWCPOS_OFFSET_MAX_SCALER = 245.9271

    # load order 1 wavecal model
    order2_metadata = get_wavecal_meta_for_spectral_order("order2")

    assert order2_metadata.order == "order2"
    assert len(order2_metadata.coefficients) == NCOEF
    assert order2_metadata.poly_degree == POLY_DEGREE
    assert order2_metadata.intercept == INTERCEPT
    assert order2_metadata.scaler_data_min_[0] == X_MIN_SCALER
    assert order2_metadata.scaler_data_max_[0] == X_MAX_SCALER
    assert order2_metadata.scaler_data_min_[1] == PWCPOS_OFFSET_MIN_SCALER
    assert order2_metadata.scaler_data_max_[1] == PWCPOS_OFFSET_MAX_SCALER


def test_load_order1_trace_model():
    # load in the trace files
    ref_trace_file = REFERENCE_TRACE_FILES["order1"]
    x, _, origin = get_reference_trace(ref_trace_file)

    # check origin is what is expected and the xlimits.
    origin = (int(origin[0]), int(origin[1]))
    x_min = min(x)
    x_max = max(x)
    x_limits = (x_min, x_max)

    assert origin == (1887, 54)
    assert x_limits == (4, 2043)


def test_load_order2_trace_model():
    # load in the trace files
    ref_trace_file = REFERENCE_TRACE_FILES["order2"]
    x, _, origin = get_reference_trace(ref_trace_file)

    # check origin is what is expected and the xlimits.
    origin = (int(origin[0]), int(origin[1]))
    x_min = min(x)
    x_max = max(x)
    x_limits = (x_min, x_max)

    assert origin == (1677, 200)
    assert x_limits == (0, 2047)


def test_rotate():
    """Test rotate step to ensure that x coordinate input is return in the
    output when interpolating back onto the pixel column grid."""

    # using trace model for rotation step
    ref_trace_file = REFERENCE_TRACE_FILES["order1"]
    x, y, origin = get_reference_trace(ref_trace_file)

    # rotate
    angle = 0.2
    x_coords, y_coords = rotate(x, y, angle, origin, interp=True)
    # checks
    assert len(x) == len(x_coords)
    assert len(y) == len(y_coords)
    assert np.array_equal(x, x_coords)



# Module to predict SOSS trace positions for a given spectral order(s)
# given the  a pupil wheel position angle taken from the "PWCPOS" fits
# header keyword.

from dataclasses import dataclass
from typing import Tuple

import numpy as np
from scipy.interpolate import interp1d
from pkg_resources import resource_filename

from pastasoss.wavecal import get_wavecal_meta_for_spectral_order
from pastasoss.wavecal import get_wavelengths

PWCPOS_CMD = 245.7600  # Commanded PWCPOS for the GR700XD


# TODO: order 3 currently unsupport ATM. Will be support in the future: TBD
REFERENCE_TRACE_FILES = {
    "order1": "jwst_niriss_gr700xd_order1_trace_refmodel.txt",
    "order2": "jwst_niriss_gr700xd_order2_trace_refmodel_002.txt",
    # order 3 currently unsupport ATM. Will be support in the future: TBD
}

REFERENCE_WAVECAL_MODELS = {
    "order1": resource_filename(
        __name__, "data/jwst_niriss_gr700xd_wavelength_model_order1.json"
    ),
    "order2": resource_filename(
        __name__, "data/jwst_niriss_gr700xd_wavelength_model_order2_002.json"
    ),
}


@dataclass
class TraceModel:
    order: str
    x: np.ndarray
    y: np.ndarray
    wavelength: np.ndarray


def rotate(
    x: np.ndarray,
    y: np.ndarray,
    angle: float,
    origin: Tuple[float, float] = (0, 0),
    interp: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Applies a rotation transformation to a set of 2D points.

    Parameters
    ----------
    x : np.ndarray
        The x-coordinates of the points to be transformed.
    y : np.ndarray
        The y-coordinates of the points to be transformed.
    angle : float
        The angle (in degrees) by which to rotate the points.
    origin : Tuple[float, float], optional
        The point about which to rotate the points. Default is (0, 0).
    interp : bool, optional
        Whether to interpolate the rotated positions onto the original x-pixel
        column values. Default is True.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The x and y coordinates of the rotated points.

    Examples
    --------
    >>> x = np.array([0, 1, 2, 3])
    >>> y = np.array([0, 1, 2, 3])
    >>> x_rot, y_rot = rotate(x, y, 90)
    """

    # shift to rotate about center
    xy_center = np.atleast_2d(origin).T
    xy = np.vstack([x, y])

    # Rotation transform matrix
    radians = np.radians(angle)
    c, s = np.cos(radians), np.sin(radians)
    R = np.array([[c, -s], [s, c]])

    # apply transformation
    x_new, y_new = R @ (xy - xy_center) + xy_center

    # interpolate rotated positions onto x-pixel column values (default)
    if interp:
        # interpolate new coordinates onto original x values and mask values
        # outside of the domain of the image 0<=x<=2047 and 0<=y<=255.
        y_new = interp1d(x_new, y_new, fill_value="extrapolate")(x)
        mask = np.where(y_new <= 255.0)
        x = x[mask]
        y_new = y_new[mask]
        return x, y_new

    return x_new, y_new


def get_reference_trace(
    file: str,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Load in the reference trace positions given a file associated with a given
    spectral order.

    Parameters
    ----------
    file : str
        The path to the file containing the reference trace positions.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        A tuple containing the x, y positions and origin of the reference
        traces.

    Examples
    --------
    >>> x, y, origin = get_reference_traces_positions('ref_filename.txt')
    """
    filepath = resource_filename(__name__, f"data/{file}")
    traces = np.loadtxt(filepath)
    origin = traces[0]
    x = traces[1:, 0]
    y = traces[1:, 1]
    return x, y, origin


def get_soss_traces(
    pwcpos: float, order: str = "123", interp: bool = True
) -> Tuple[np.ndarray, np.ndarray]:
    """
    This is the primary method for generate the gr700xd trace position given a
    pupil wheel positions angle provided in the FITS header under keyword
    PWCPOS. The traces for a given spectral order are found by perform a
    rotation transformation using the refence trace positions at the commanded
    PWCPOS=245.76 degrees. This methods yield sub-pixel performance and will be
    improved upon in later interations as more NIRISS/SOSS observations become
    available.

    Parameters
    ----------
    pwcpos : float
        The pupil wheel positions angle provided in the FITS header under
        keyword PWCPOS.
    order : str, optional
        The spectral order to compute the new traces for. Default is '123'.
        Support for order 3 will be added at a later date.
    interp : bool, optional
        Whether to interpolate the rotated positions onto the original x-pixel
        column values. Default is True.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]]
    If `order` is '1', a tuple of the x and y coordinates of the rotated
    points for the first spectral order.
    If `order` is '2', a tuple of the x and y coordinates of the rotated
    points for the second spectral order.
    If `order` is '3' or a combination of '1', '2', and '3', a list of
    tuples of the x and y coordinates of the rotated points for each
    spectral order.

    Raises
    ------
    ValueError
        If `order` is not '1', '2', '3', or a combination of '1', '2', and '3'.

    Examples
    --------
    >>> x_new, y_new = get_trace_from_reference_transform(2.3)
    """

    norders = len(order)

    if norders > 3:
        raise ValueError("order must be: 1,2, 3.")
    if norders > 1:
        # recursively compute the new traces for each order
        return [get_soss_traces(pwcpos, m) for m in order]

    # This might be an alternative way of writing this
    # if 'order'+order in REFERENCE_TRACE_FILES.keys():
    #     ref_file = REFERENCE_TRACE_FILES["order"+order]
    # This section can definite be refactored in a later version
    elif order == "1":
        ref_trace_file = REFERENCE_TRACE_FILES["order1"]
        wave_cal_model_meta = get_wavecal_meta_for_spectral_order("order1")

    elif order == "2":
        ref_trace_file = REFERENCE_TRACE_FILES["order2"]
        wave_cal_model_meta = get_wavecal_meta_for_spectral_order("order2")

    elif order == "3":
        print("The software currently does not support order 3 at this time.")
        return None

    # reference trace data
    x, y, origin = get_reference_trace(ref_trace_file)

    # rotated reference trace
    x_new, y_new = rotate(x, y, pwcpos - PWCPOS_CMD, origin, interp=interp)

    # wavelength associated to trace at given pwcpos value
    wavelengths = get_wavelengths(x_new, pwcpos, wave_cal_model_meta)

    # return x_new, y_new, wavelengths
    return TraceModel(order, x_new, y_new, wavelengths)



# Wavecal module for handling the wavelength calibration model. For now we
# will use the reference json files that has all the model information. Ideally
# we would like to use just the train model for production but for now this is
# this is fine.

import json
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict

import numpy as np
import numpy.typing as npt
from pkg_resources import resource_filename

# explicitly defining the commanded position here as well (this is temp)
PWCPOS_CMD = 245.76

# TODO: order 3 currently unsupport ATM. Will be support in the future: TBD
REFERENCE_WAVECAL_MODELS = {
    "order1": resource_filename(
        __name__, "data/jwst_niriss_gr700xd_wavelength_model_order1.json"
    ),
    "order2": resource_filename(
        __name__, "data/jwst_niriss_gr700xd_wavelength_model_order2_002.json"
    ),
}


# dataclass object to store the metadata from a wavecal model
@dataclass
class NIRISS_GR700XD_WAVECAL_META:
    """
    Datamodel object container for the wavecal meta data in the json reference
    file.
    """

    order: str
    coefficients: npt.NDArray[np.float64]
    intercept: npt.NDArray[np.float64]
    scaler_data_min_: float
    scaler_data_max_: float
    poly_degree: int


def load_wavecal_model(filename: str) -> Dict[str, Any]:
    """
    Load a wavecalibration model from a JSON file.

    Parameters
    ----------
    filename : str
        The path to the JSON file containing the wavecalibration model.

    Returns
    -------
    dict
        A dictionary representing the loaded wavecalibration model.

    Notes
    -----
    This function reads the specified JSON file and returns its contents as a
    dictionary, which typically contains information about a wavecalibration
    model used in data analysis.
    """
    with open(filename, "r") as file:
        wavecal_model = json.load(file)
    return wavecal_model


def get_wavecal_meta_for_spectral_order(
    order: str,
) -> NIRISS_GR700XD_WAVECAL_META:
    """
    Retrieve wavecalibration model metadata for a specific spectral order.

    Parameters
    ----------
    order : str
        The spectral order for which wavecalibration metadata is requested.
        Valid options are 'order 1', 'order 2', or 'order 3'.

    Returns
    -------
    NIRISS_GR700XD_WAVECAL_META
        An object containing wavecalibration metadata, including order, model
        coefficients, intercept, and scaler data bounds.

    Raises
    ------
    ValueError
        If the provided 'order' is not one of the valid options.

    Notes
    -----
    This function retrieves wavecalibration metadata for the specified spectral
    order. It first checks if the 'order' is valid and then loads the
    corresponding wavecalibration model. The model's coefficients, intercept,
    and scaler data bounds are extracted and returned as part of the metadata
    object.
    """
    # get the reference wavecal file name
    if order not in REFERENCE_WAVECAL_MODELS:
        raise ValueError(f"valid orders are: {REFERENCE_WAVECAL_MODELS.keys()}.")

    # get the appropiate reference file name given the order
    reference_filename = REFERENCE_WAVECAL_MODELS[order]

    # load in the model
    wavecal_model = load_wavecal_model(reference_filename)

    # model coefficients
    poly_degree = wavecal_model["model"]["poly_deg"]
    coefficients = wavecal_model["model"]["coef"]
    intercept = wavecal_model["model"]["intercept"]

    # info for scaling inputs
    scaler_data_min_ = wavecal_model["model"]["scaler"]["data_min_"]
    scaler_data_max_ = wavecal_model["model"]["scaler"]["data_max_"]

    return NIRISS_GR700XD_WAVECAL_META(
        order, coefficients, intercept, scaler_data_min_, scaler_data_max_, poly_degree
    )


def get_wavelengths(
    x: np.ndarray, pwcpos: float, wavecal_meta: NIRISS_GR700XD_WAVECAL_META
) -> np.ndarray:
    """Get the associated wavelength values for a given spectral order"""
    if wavecal_meta.order == "order1":
        wavelengths = wavecal_model_order1_poly(x, pwcpos, wavecal_meta)
    elif wavecal_meta.order == "order2":
        # raise NotImplementedError("Order 2 not implemented")
        wavelengths = wavecal_model_order2_poly(x, pwcpos, wavecal_meta)
    elif wavecal_meta.order == "order3":
        raise ValueError("Order 3 not supported at this time")
    else:
        raise ValueError("not a valid order")

    return wavelengths


def min_max_scaler(x, x_min, x_max):
    """
    Apply min-max scaling to input values.

    Parameters
    ----------
    x : float or numpy.ndarray
        The input value(s) to be scaled.
    x_min : float
        The minimum value in the range to which 'x' will be scaled.
    x_max : float
        The maximum value in the range to which 'x' will be scaled.

    Returns
    -------
    float or numpy.ndarray
        The scaled value(s) in the range [0, 1].

    Notes
    -----
    Min-max scaling is a data normalization technique that scales input values
    'x' to the range [0, 1] based on the provided minimum and maximum values,
    'x_min' and 'x_max'. This function is applicable to both individual values
    and arrays of values. This function will use the min/max values from the
    training data of the wavecal model.
    """
    # scaling the input x values
    x_scaled = (x - x_min) / (x_max - x_min)
    return x_scaled


def wavecal_model_order1_poly(x, pwcpos, wavecal_meta: NIRISS_GR700XD_WAVECAL_META):
    """compute order 1 wavelengths"""
    x_scaler = partial(
        min_max_scaler,
        **{
            "x_min": wavecal_meta.scaler_data_min_[0],
            "x_max": wavecal_meta.scaler_data_max_[0],
        },
    )

    pwcpos_offset_scaler = partial(
        min_max_scaler,
        **{
            "x_min": wavecal_meta.scaler_data_min_[1],
            "x_max": wavecal_meta.scaler_data_max_[1],
        },
    )

    def get_poly_features(x: np.array, offset: np.array) -> np.ndarray:
        """polynomial features for the order 1 wavecal model"""
        poly_features = np.array(
            [
                x,
                offset,
                x**2,
                x * offset,
                offset**2,
                x**3,
                x**2 * offset,
                x * offset**2,
                offset**3,
                x**4,
                x**3 * offset,
                x**2 * offset**2,
                x * offset**3,
                offset**4,
                x**5,
                x**4 * offset,
                x**3 * offset**2,
                x**2 * offset**3,
                x * offset**4,
                offset**5,
            ]
        )
        return poly_features

    # extract model weights and intercept
    coef = wavecal_meta.coefficients
    intercept = wavecal_meta.intercept

    # get pixel columns and then scaled
    x_scaled = x_scaler(x)

    # offset
    offset = np.ones_like(x) * (pwcpos - PWCPOS_CMD)
    offset_scaled = pwcpos_offset_scaler(offset)

    # polynomial features
    poly_features = get_poly_features(x_scaled, offset_scaled)
    wavelengths = coef @ poly_features + intercept

    return wavelengths


def wavecal_model_order2_poly(x, pwcpos, wavecal_meta: NIRISS_GR700XD_WAVECAL_META):
    """compute order 2 wavelengths"""
    x_scaler = partial(
        min_max_scaler,
        **{
            "x_min": wavecal_meta.scaler_data_min_[0],
            "x_max": wavecal_meta.scaler_data_max_[0],
        },
    )

    pwcpos_offset_scaler = partial(
        min_max_scaler,
        **{
            "x_min": wavecal_meta.scaler_data_min_[1],
            "x_max": wavecal_meta.scaler_data_max_[1],
        },
    )

    def get_poly_features(x: np.array, offset: np.array) -> np.ndarray:
        """Polynomial features for the order 2 wavecal model"""
        poly_features = np.array(
            [
                x,
                offset,
                x**2,
                x * offset,
                offset**2,
                x**3,
                x**2 * offset,
                x * offset**2,
                offset**3,
            ]
        )
        return poly_features

    # coef and intercept
    coef = wavecal_meta.coefficients
    intercept = wavecal_meta.intercept

    # get pixel columns and then scaled
    x_scaled = x_scaler(x)

    # offset
    # offset = np.ones_like(x) * (pwcpos - PWCPOS_CMD)
    # this will need to get changed later...
    offset = np.ones_like(x) * pwcpos
    offset_scaled = pwcpos_offset_scaler(offset)

    # polynomial features
    poly_features = get_poly_features(x_scaled, offset_scaled)
    wavelengths = coef @ poly_features + intercept

    return wavelengths
